@@ -215,11 +215,11 @@ index_add_default_filters()
 			 false));
     index_command("text/vcard",
 		  Filter(get_pkglibbindir() + "/vcard2text", false));
-    index_command("application/vnd.apply.keynote",
+    index_command("application/vnd.apple.keynote",
 		  Filter("key2text", false));
-    index_command("application/vnd.apply.numbers",
+    index_command("application/vnd.apple.numbers",
 		  Filter("numbers2text", false));
-    index_command("application/vnd.apply.pages",
+    index_command("application/vnd.apple.pages",
 		  Filter("pages2text", false));
 }
 
@@ -622,11 +622,11 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 			substituted = true;
 			string tail(cmd, pcent + 2);
 			cmd.resize(pcent);
-			append_filename_argument(cmd, file);
-			// Remove the space append_filename_argument() adds before
-			// the argument - the command string either includes one,
-			// or won't expect one (e.g. --input=%f).
-			cmd.erase(pcent, 1);
+			// Suppress the space append_filename_argument()
+			// usually adds before the argument - the command
+			// string either includes one, or won't expect one
+			// (e.g. --input=%f).
+			append_filename_argument(cmd, file, false);
 			pcent = cmd.size();
 			cmd += tail;
 			break;
@@ -647,11 +647,11 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 			substituted = true;
 			string tail(cmd, pcent + 2);
 			cmd.resize(pcent);
-			append_filename_argument(cmd, tmpout);
-			// Remove the space append_filename_argument() adds before
-			// the argument - the command string either includes one,
-			// or won't expect one (e.g. --input=%f).
-			cmd.erase(pcent, 1);
+			// Suppress the space append_filename_argument()
+			// usually adds before the argument - the command
+			// string either includes one, or won't expect one
+			// (e.g. --output=%t).
+			append_filename_argument(cmd, tmpout, false);
 			pcent = cmd.size();
 			cmd += tail;
 			break;
@@ -796,11 +796,10 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 		     d.get_size(), d.get_mtime());
 		return;
 	    }
-	    string cmd = "ps2pdf";
-	    append_filename_argument(cmd, file);
+	    string cmd = "ps2pdf -";
 	    append_filename_argument(cmd, tmpfile);
 	    try {
-		run_filter(cmd, false);
+		run_filter(d.get_fd(), cmd, false);
 		cmd = "pdftotext -enc UTF-8";
 		append_filename_argument(cmd, tmpfile);
 		cmd += " -";
