@@ -100,6 +100,9 @@ static int my_snprintf(char *str, size_t size, const char *format, ...)
 #define my_snprintf SNPRINTF
 #endif
 
+/// Map shard to DB parameter value and stats to allow docid mapping.
+vector<SubDB> subdbs;
+
 static bool query_parsed = false;
 static bool done_query = false;
 static Xapian::docid last = 0;
@@ -1040,16 +1043,16 @@ struct func_desc {
     struct func_attrib a;
 };
 
-#define N -1
+#define N (-1)
 #define M 'M'
 #define Q 'Q'
 // NB when adding a new command which ensures M or Q, update the list in
 // docs/omegascript.rst
-static struct func_desc func_tab[] = {
+static const struct func_desc func_tab[] = {
 //name minargs maxargs evalargs ensure
 {"",{CMD_,	   N, N, 0, 0}},// commented out code
 T(add,		   0, N, N, 0), // add a list of numbers
-T(addfilter,	   1, 1, N, 0), // add filter term
+T(addfilter,	   1, 2, N, 0), // add filter term
 T(allterms,	   0, 1, N, 0), // list of all terms matching document
 T(and,		   1, N, 0, 0), // logical shortcutting and of a list of values
 T(cgi,		   1, 1, N, 0), // return cgi parameter value
@@ -1057,7 +1060,7 @@ T(cgilist,	   1, 1, N, 0), // return list of values for cgi parameter
 T(cgiparams,	   0, 0, N, 0), // return list of cgi parameter names
 T(chr,		   1, 1, N, 0), // return UTF-8 for given Unicode codepoint
 T(collapsed,	   0, 0, N, 0), // return number of hits collapsed into this
-T(cond,		   2, N, 0, 0), // return position of substring, or empty string
+T(cond,		   2, N, 0, 0), // cascaded conditionals
 T(contains,	   2, 2, N, 0), // return position of substring, or empty string
 T(csv,		   1, 2, N, 0), // CSV string escaping
 T(date,		   1, 2, N, 0), // convert time_t to strftime format
@@ -1138,7 +1141,7 @@ T(setmap,	   1, N, N, 0), // set map of option values
 T(setrelevant,	   0, 1, N, Q), // set rset
 T(slice,	   2, 2, N, 0), // slice a list using a second list
 T(snippet,	   1, 2, N, M), // generate snippet from text
-T(sort,		   1, 2, N, M), // alpha sort a list
+T(sort,		   1, 2, N, 0), // alpha sort a list
 T(split,	   1, 2, N, 0), // split a string to give a list
 T(stoplist,	   0, 0, N, Q), // return list of stopped terms
 T(sub,		   2, 2, N, 0), // subtract
@@ -1192,29 +1195,12 @@ write_all(int fd, const char * buf, size_t count)
     return 0;
 }
 
-static const vector<string>&
-get_subdbs()
-{
-    static vector<string> subdbs;
-    if (subdbs.empty()) {
-	size_t p = 0, q;
-	while (true) {
-	    q = dbname.find('/', p);
-	    subdbs.emplace_back(dbname, p, q - p);
-	    if (q == string::npos) break;
-	    p = q + 1;
-	}
-    }
-    return subdbs;
-}
-
 static string
 eval(const string &fmt, const vector<string> &param)
 {
     static map<string, const struct func_attrib *> func_map;
     if (func_map.empty()) {
-	struct func_desc *p;
-	for (p = func_tab; p->name != NULL; ++p) {
+	for (auto p = func_tab; p->name != NULL; ++p) {
 	    func_map[string(p->name)] = &(p->a);
 	}
     }
@@ -1335,7 +1321,16 @@ eval(const string &fmt, const vector<string> &param)
 		break;
 	    }
 	    case CMD_addfilter:
-		add_bterm(args[0]);
+		if (args.size() == 1 || args[1].empty() || args[1] == "B") {
+		    add_bterm(args[0]);
+		} else if (args[1] == "N") {
+		    add_nterm(args[0]);
+		} else {
+		    string msg = "Invalid $addfilter type '";
+		    msg += args[1];
+		    msg += "'";
+		    throw msg;
+		}
 		break;
 	    case CMD_allterms: {
 		// list of all terms indexing document
@@ -1634,7 +1629,7 @@ eval(const string &fmt, const vector<string> &param)
 		// 0-based mset index
 		value = str(hit_no);
 		break;
-	    case CMD_hitlist:
+	    case CMD_hitlist: {
 #if 0
 		url_query_string = "?DB=";
 		url_query_string += dbname;
@@ -1650,22 +1645,22 @@ eval(const string &fmt, const vector<string> &param)
 		    int ch;
 		    while ((ch = *q++) != '\0') {
 			switch (ch) {
-			 case '+':
+			  case '+':
 			    url_query_string += "%2b";
 			    break;
-			 case '"':
+			  case '"':
 			    url_query_string += "%22";
 			    break;
-			 case '%':
+			  case '%':
 			    url_query_string += "%25";
 			    break;
-			 case '&':
+			  case '&':
 			    url_query_string += "%26";
 			    break;
-			 case ' ':
+			  case ' ':
 			    ch = '+';
 			    /* fall through */
-			 default:
+			  default:
 			    url_query_string += ch;
 			}
 		    }
@@ -1676,10 +1671,12 @@ eval(const string &fmt, const vector<string> &param)
 		    url_query_string += i->second;
 		}
 #endif
+		auto save_hit_no = hit_no;
 		for (hit_no = topdoc; hit_no < last; ++hit_no)
 		    value += print_caption(args[0], param);
-		hit_no = 0;
+		hit_no = save_hit_no;
 		break;
+	    }
 	    case CMD_hitsperpage:
 		value = str(hits_per_page);
 		break;
@@ -1892,6 +1889,15 @@ eval(const string &fmt, const vector<string> &param)
 		value = str(val);
 		break;
 	    }
+	    case CMD_mod: {
+		int denom = string_to_int(args[1]);
+		if (denom == 0) {
+		    value = "divide by 0";
+		} else {
+		    value = str(string_to_int(args[0]) % denom);
+		}
+		break;
+	    }
 	    case CMD_msize:
 		// Estimated number of matches.
 		value = str(mset.get_matches_estimated());
@@ -1910,15 +1916,6 @@ eval(const string &fmt, const vector<string> &param)
 		// Upper bound on number of matches.
 		value = str(mset.get_matches_upper_bound());
 		break;
-	    case CMD_mod: {
-		int denom = string_to_int(args[1]);
-		if (denom == 0) {
-		    value = "divide by 0";
-		} else {
-		    value = str(string_to_int(args[0]) % denom);
-		}
-		break;
-	    }
 	    case CMD_mul: {
 		vector<string>::const_iterator i = args.begin();
 		int total = string_to_int(*i++);
@@ -2162,14 +2159,19 @@ eval(const string &fmt, const vector<string> &param)
 	    case CMD_subdb: {
 		Xapian::docid id = q0;
 		if (args.size() > 0) id = string_to_int(args[0]);
-		auto subdbs = get_subdbs();
-		value = subdbs[(id - 1) % subdbs.size()];
+		value = subdbs[(id - 1) % subdbs.size()].get_name();
 		break;
 	    }
 	    case CMD_subid: {
 		Xapian::docid id = q0;
 		if (args.size() > 0) id = string_to_int(args[0]);
-		value = str(((id - 1) / get_subdbs().size()) + 1);
+		// This is the docid in the single shard.
+		Xapian::docid shard_did = (id - 1) / subdbs.size() + 1;
+		// We now need to map this back to the docid in the collection
+		// of shards specified by the DB parameter value which $subdb
+		// returns.
+		const SubDB& subdb = subdbs[(id - 1) % subdbs.size()];
+		value = str(subdb.map_docid(shard_did));
 		break;
 	    }
 	    case CMD_substr: {
