@@ -45,6 +45,8 @@
 #include "hashterm.h"
 #include "loadfile.h"
 #include "myhtmlparse.h"
+#include "parseint.h"
+#include "setenv.h"
 #include "str.h"
 #include "stringutils.h"
 #include "timegm.h"
@@ -87,21 +89,21 @@ const char * action_names[] = {
 #define DUMP_ACTION(A) cout << action_names[(A).get_action()] << "(" << (A).get_string_arg() << "," << (A).get_num_arg() << ")" << endl
 
 class Action {
-public:
+  public:
     typedef enum {
 	BAD, NEW,
 	BOOLEAN, DATE, FIELD, HASH, HEXTOBIN, INDEX, INDEXNOPOS, LOAD, LOWER,
 	PARSEDATE, SPELL, SPLIT, TRUNCATE, UNHTML, UNIQUE, VALUE,
 	VALUENUMERIC, VALUEPACKED, WEIGHT
     } type;
-    enum { SPLIT_NONE, SPLIT_DEDUP, SPLIT_SORT };
-private:
+    enum { SPLIT_NONE, SPLIT_DEDUP, SPLIT_SORT, SPLIT_PREFIXES };
+  private:
     type action;
     int num_arg;
     string string_arg;
     // Offset into indexscript line.
     size_t pos;
-public:
+  public:
     Action(type action_, size_t pos_)
 	: action(action_), num_arg(0), pos(pos_) { }
     Action(type action_, size_t pos_, const string & arg)
@@ -484,6 +486,17 @@ bad_escaping:
 		    }
 		}
 		switch (code) {
+		    case Action::DATE:
+			if (val != "unix" &&
+			    val != "unixutc" &&
+			    val != "yyyymmdd") {
+			    report_location(DIAG_ERROR, filename, line_no);
+			    cerr << "Invalid parameter '" << val << "' for "
+				    "action 'date'" << endl;
+			    exit(1);
+			}
+			actions.emplace_back(code, action_pos, val);
+			break;
 		    case Action::INDEX:
 		    case Action::INDEXNOPOS:
 			actions.emplace_back(code, action_pos, val, weight);
@@ -500,6 +513,23 @@ bad_escaping:
 			}
 			useless_weight_pos = action_pos;
 			break;
+		    case Action::PARSEDATE: {
+			if (val.find("%Z") != val.npos) {
+			    report_location(DIAG_ERROR, filename, line_no);
+			    cerr << "Parsing timezone names with %Z is not supported" << endl;
+			    exit(1);
+			}
+#ifndef HAVE_STRUCT_TM_TM_GMTOFF
+			if (val.find("%z") != val.npos) {
+			    report_location(DIAG_ERROR, filename, line_no);
+			    cerr << "Parsing timezone offsets with %z is not supported on "
+				    "this platform" << endl;
+			    exit(1);
+			}
+#endif
+			actions.emplace_back(code, action_pos, val);
+			break;
+		    }
 		    case Action::SPLIT: {
 			if (val.empty()) {
 			    report_location(DIAG_ERROR, filename, line_no);
@@ -514,6 +544,8 @@ bad_escaping:
 				operation = Action::SPLIT_SORT;
 			    } else if (vals[1] == "none") {
 				operation = Action::SPLIT_NONE;
+			    } else if (vals[1] == "prefixes") {
+				operation = Action::SPLIT_PREFIXES;
 			    } else {
 				report_location(DIAG_ERROR, filename, line_no);
 				cerr << "Bad split operation '" << vals[1]
@@ -792,14 +824,15 @@ badhex:
 		    ++split_end;
 		}
 
+		int split_type = action.get_num_arg();
 		if (value.empty()) {
 		    // Nothing to do.
-		} else if (action.get_num_arg() != Action::SPLIT_SORT) {
+		} else if (split_type != Action::SPLIT_SORT) {
 		    // Generate split as we consume it.
 		    const string& delimiter = action.get_string_arg();
 
 		    unique_ptr<unordered_set<string>> seen;
-		    if (action.get_num_arg() == Action::SPLIT_DEDUP) {
+		    if (split_type == Action::SPLIT_DEDUP) {
 			seen.reset(new unordered_set<string>);
 		    }
 
@@ -809,7 +842,18 @@ badhex:
 			string::size_type i = 0;
 			while (true) {
 			    string::size_type j = value.find(ch, i);
-			    if (i != j) {
+			    if (split_type == Action::SPLIT_PREFIXES) {
+				if (j > 0) {
+				    string val(value, 0, j);
+				    run_actions(action_it, split_end,
+						database, indexer,
+						val,
+						this_field_is_content, doc,
+						fields,
+						field, fname, line_no,
+						docid);
+				}
+			    } else if (i != j) {
 				string val(value, i, j - i);
 				if (!seen.get() || seen->insert(val).second) {
 				    run_actions(action_it, split_end,
@@ -828,7 +872,18 @@ badhex:
 			string::size_type i = 0;
 			while (true) {
 			    string::size_type j = value.find(delimiter, i);
-			    if (i != j) {
+			    if (split_type == Action::SPLIT_PREFIXES) {
+				if (j > 0) {
+				    string val(value, 0, j);
+				    run_actions(action_it, split_end,
+						database, indexer,
+						val,
+						this_field_is_content, doc,
+						fields,
+						field, fname, line_no,
+						docid);
+				}
+			    } else if (i != j) {
 				string val(value, i, j - i);
 				if (!seen.get() || seen->insert(val).second) {
 				    run_actions(action_it, split_end,
@@ -983,13 +1038,36 @@ badhex:
 		const string & type = action.get_string_arg();
 		string yyyymmdd;
 		if (type == "unix") {
-		    time_t t = atoi(value.c_str());
+		    time_t t;
+		    if (!parse_signed(value.c_str(), t)) {
+			report_location(DIAG_WARN, fname, line_no);
+			cerr << "Date value (in secs) for action DATE "
+				"must be an integer - ignoring" << endl;
+			break;
+		    }
 		    struct tm *tm = localtime(&t);
 		    int y = tm->tm_year + 1900;
 		    int m = tm->tm_mon + 1;
 		    yyyymmdd = date_to_string(y, m, tm->tm_mday);
+		} else if (type == "unixutc") {
+		    time_t t;
+		    if (!parse_signed(value.c_str(), t)) {
+			report_location(DIAG_WARN, fname, line_no);
+			cerr << "Date value (in secs) for action DATE "
+				"must be an integer - ignoring" << endl;
+			break;
+		    }
+		    struct tm *tm = gmtime(&t);
+		    int y = tm->tm_year + 1900;
+		    int m = tm->tm_mon + 1;
+		    yyyymmdd = date_to_string(y, m, tm->tm_mday);
 		} else if (type == "yyyymmdd") {
-		    if (value.length() != 8) break;
+		    if (value.length() != 8) {
+			report_location(DIAG_WARN, fname, line_no);
+			cerr << "date=yyyymmdd expects an 8 character value "
+				"- ignoring" << endl;
+			break;
+		    }
 		    yyyymmdd = value;
 		}
 
@@ -1022,8 +1100,14 @@ badhex:
 			    "(\"" << ret << "\" left over) but "
 			    "indexing anyway" << endl;
 		}
-
-		value = str(timegm(&tm));
+#ifdef HAVE_STRUCT_TM_TM_GMTOFF
+		auto gmtoff = tm.tm_gmtoff;
+#endif
+		auto secs_since_epoch = timegm(&tm);
+#ifdef HAVE_STRUCT_TM_TM_GMTOFF
+		secs_since_epoch -= gmtoff;
+#endif
+		value = str(secs_since_epoch);
 		break;
 	    }
 	    default:
@@ -1162,6 +1246,10 @@ try {
     verbose = false;
     Xapian::Stem stemmer("english");
 
+    // Without this, strptime() seems to treat formats without a timezone as
+    // being local time, including %s.
+    setenv("TZ", "UTC", 1);
+
     constexpr auto NO_ARG = no_argument;
     constexpr auto REQ_ARG = required_argument;
     static const struct option longopts[] = {
